"""
Local flux solve in the vicinity of a collection of filaments with polygon sections
that do not necessarily fall on a rectangular grid by allocating the fraction of
intersecting polygon area to each mesh cell.
"""

from dataclasses import dataclass
from itertools import chain

import numpy as np
from interpn import MulticubicRegular
from numpy.typing import NDArray
from pytest import approx
from scipy.optimize import fsolve
from shapely import GeometryCollection, MultiPolygon, Polygon

from device_inductance.logging import log
from device_inductance.utils import calc_flux_density_from_flux, solve_flux_axisymmetric

RMIN = 2e-2
"""[m] minimum r-value to allow in local solver grid"""

N_INSIDE = 8
"""Target number of grid points inside filament extent"""

N_OUTSIDE = 9
"""Target number of grid points on either side of filament extent"""


@dataclass(frozen=True)
class LocalFields:
    """Outputs of local field solve"""

    jtor_per_amp: NDArray
    """[1/m^2] normalized current density map"""

    psi_per_amp: NDArray
    """[Wb/A] normalized poloidal flux"""

    br_per_amp: NDArray
    """[T/A] normalized flux density, r-component"""

    bz_per_amp: NDArray
    """[T/A] normalized flux density, z-component"""

    psi_per_amp_interpolator: MulticubicRegular
    """[Wb/A] normalized poloidal flux, hermite spline interpolator"""

    br_per_amp_interpolator: MulticubicRegular
    """[T/A] normalized flux density, r-component, hermite spline interpolator"""

    bz_per_amp_interpolator: MulticubicRegular
    """[T/A] normalized flux density, z-component, hermite spline interpolator"""

    grids: tuple[NDArray, NDArray]
    """[m] 1D computational grids"""

    meshes: tuple[NDArray, NDArray]
    """[m] 2D computational meshgrids"""


def local_fields(fil_rzn: tuple[NDArray, NDArray, NDArray], polygons: list[Polygon]) -> LocalFields:
    """
    Estimate local self-field of a collection of filaments with polygon representations.

    Args:
        fil_rzn: r-coord, z-coord, and number of turns for each filament
        polygons: polygon section of each filament

    Returns:
        Structure containing local field map components and interpolators
    """
    rs, zs, ns = fil_rzn
    grids_regular = _make_grids_regular(rs, zs)

    # If the winding pack lands on a regular grid, use the corresponding regular grid.
    # Otherwise, fall back on a heuristic grid that spans the winding pack.
    grids = grids_regular or _make_grids_irregular(polygons)

    meshes = _make_mesh(grids)
    dr = grids[0][1] - grids[0][0]
    dz = grids[1][1] - grids[1][0]
    dxgrid = (dr, dz)

    if grids_regular is not None:
        # Just put all the current for each filament on its corresponding mesh cell
        jtor_per_amp = _allocate_current_regular(fil_rzn, grids, meshes)
    else:
        # Split polygons between mesh cells, run flux solve, and back-out B-fields
        jtor_per_amp = _allocate_current_irregular(fil_rzn, polygons, meshes, dxgrid)

    psi_per_amp, br_per_amp, bz_per_amp = _local_fields(jtor_per_amp, grids, meshes)

    # Make interpolators
    dims = [len(grids[0]), len(grids[1])]
    starts = np.array([grids[0][0], grids[1][0]])
    steps = np.array(dxgrid)
    psi_per_amp_interpolator = MulticubicRegular.new(dims, starts, steps, psi_per_amp.flatten())
    br_per_amp_interpolator = MulticubicRegular.new(dims, starts, steps, br_per_amp.flatten())
    bz_per_amp_interpolator = MulticubicRegular.new(dims, starts, steps, bz_per_amp.flatten())

    # Pack numerous outputs into a struct
    result = LocalFields(
        jtor_per_amp,
        psi_per_amp,
        br_per_amp,
        bz_per_amp,
        psi_per_amp_interpolator,
        br_per_amp_interpolator,
        bz_per_amp_interpolator,
        grids,
        meshes,
    )

    return result


def _filament_extent(polygons: list[Polygon]) -> tuple[float, float, float, float]:
    """Get the (rmin, rmax, zmin, zmax) extent that bounds the filament polygons."""
    r = np.array([*chain([p.boundary.xy[0] for p in polygons])])
    z = np.array([*chain([p.boundary.xy[1] for p in polygons])])

    rmin, rmax = np.min(r), np.max(r)
    zmin, zmax = np.min(z), np.max(z)
    extent = (float(rmin), float(rmax), float(zmin), float(zmax))

    return extent


def _make_grids_irregular(polygons: list[Polygon]) -> tuple[NDArray, NDArray]:
    """Make grids that bound the filament extent
    plus at least 7 cells outside to support a 4th order difference method."""
    rmin, rmax, zmin, zmax = _filament_extent(polygons)

    if rmin < 0.0:
        raise ValueError(
            f"Minimum r-coordinate of a filament is <{RMIN} [m],"
            " which does not leave enough room to set boundary conditions."
        )
    if rmin < RMIN:
        log().warning(
            f"Clamping filament polygon bounds to minimum radius of {RMIN} [m], which may distort fields."
        )
        rmin = RMIN

    rspan = rmax - rmin
    zspan = zmax - zmin

    # This is the grid step size if there is no conflict with crossing R=0
    n_inside = max(int(float(len(polygons) ** 0.5)), N_INSIDE)
    dr = rspan / n_inside
    dz = zspan / n_inside

    # Take the smaller step size of the two for both;
    # the method can handle different step sizes on each axis, but best
    # results are obtained with grid cells of equal aspect ratio.
    dr = min(dr, dz)
    dz = dr

    # Check if we are interfering with R=0
    # and decrease size of both r-step and z-step until that is resolved
    # to keep the innermost point off of R=0.
    while rmin - dr * N_OUTSIDE < RMIN / 2.0:
        dr /= 2.0
    dz = dr

    # Make the actual grids
    rgrid = np.arange(rmin - dr * N_OUTSIDE, rmax + dr * N_OUTSIDE + dr, dr)
    zgrid = np.arange(zmin - dz * N_OUTSIDE, zmax + dz * N_OUTSIDE + dz, dz)

    nr, nz = len(rgrid), len(zgrid)
    if nz > 100 or nz > 100:
        log().warning(
            f"Using an excessively large mesh ({nr} X {nz}) "
            "to represent filament local flux solve due to proximity to R=0."
        )

    return rgrid, zgrid


def _make_mesh(grids: tuple[NDArray, NDArray]) -> tuple[NDArray, NDArray]:
    rmesh, zmesh = np.meshgrid(*grids, indexing="ij")
    return rmesh, zmesh


def _allocate_current_irregular(
    fil_rzn: tuple[NDArray, NDArray, NDArray],
    fil_polygons: list[Polygon],
    meshes: tuple[NDArray, NDArray],
    dxgrid: tuple[float, float],
) -> NDArray:
    """Convert a collection of filaments with polygon
    representations to a jtor array mapped on to the meshes."""
    rs, zs, ns = fil_rzn
    total_turns = np.sum(ns)
    dr, dz = dxgrid
    rmesh, zmesh = meshes

    # Make polygons for each mesh cell
    def cell_to_poly(r, z) -> Polygon:
        """Rectangular polygon representing a mesh cell"""
        return Polygon.from_bounds(r - dr / 2, z - dz / 2, r + dr / 2, z + dz / 2)

    mesh_polygons = [cell_to_poly(r, z) for r, z in zip(rmesh.flatten(), zmesh.flatten(), strict=True)]

    # For each filament polygon, find the fraction of its area
    # that falls in each mesh cell polygon.
    # The fraction of the total area assigned to each filament
    # should be already accounted in the `n` for the filament.
    itor_per_amp = np.zeros_like(rmesh.flatten())
    for i in range(len(itor_per_amp)):
        mp = mesh_polygons[i]
        for n, fp in zip(ns, fil_polygons, strict=True):
            # Find overlapping area between this filament and this grid cell.
            # If the filament polygon is not well-behaved, we can end up with more
            # than one distinct overlapping region.
            intersection = fp.intersection(mp)
            if isinstance(intersection, Polygon):
                # Simple intersection
                overlap_area = intersection.area
            elif isinstance(intersection, GeometryCollection | MultiPolygon):
                # Non-simple intersection
                g = intersection.geoms
                overlap_area = sum([x.area for x in g if isinstance(x, Polygon)])
            else:
                raise TypeError(f"Unexpected intersection result: {intersection}")

            # Add contribution to this grid cell from this filament
            itor_per_amp[i] += n * overlap_area / fp.area

    assert np.sum(itor_per_amp) == approx(total_turns, rel=1e-6, abs=1e-6)
    itor_per_amp = itor_per_amp.reshape(rmesh.shape)

    # Shift the centroid of the current map to match the centroid of the filaments
    # Otherwise, there will be a slight discontinuity between the local field map and the
    # far-field values calculated from direct filament functions.
    rcfil = np.sum(rs * ns) / total_turns  # [m]
    zcfil = np.sum(zs * ns) / total_turns
    fil_centroid = np.array((rcfil, zcfil))

    def shift2d(v, shift):
        """Shift a fraction of each cell's current to its neighbor in a 2D array"""
        rshift, zshift = shift
        nr, nz = v.shape

        # Make linear basis functions about the center of the array
        rinds = np.array(range(nr), dtype=float)  # Indices as coordinate
        zinds = np.array(range(nz), dtype=float)
        rmid = (nr - 1) / 2  # Normalization factor to map coordinates to [-1, 1]
        zmid = (nz - 1) / 2
        rcoord = (rinds - rmid) / rmid  # Normalized array coordinates
        zcoord = (zinds - zmid) / zmid
        rbasis = rshift * rcoord  # Shift basis functions
        zbasis = zshift * zcoord

        # Broadcast shift bases
        vshifted = v + v * rbasis.reshape(nr, 1) + v * zbasis.reshape(1, nz)

        # Rescale to preserve total
        vshifted *= np.sum(v) / np.sum(vshifted)

        return vshifted

    def itor_centroid(shift):
        """Find the mesh current centroid given some percent shift in current between neighboring cells"""
        itor_shifted = shift2d(itor_per_amp, shift)
        isum = np.sum(itor_shifted)
        rc = np.sum(itor_shifted * rmesh) / isum
        zc = np.sum(itor_shifted * zmesh) / isum

        return np.array((rc, zc))

    # Solve a shift to match the current centroid of the mesh representation to the
    # current centroid of the filament representation.
    shift = fsolve(lambda shift: itor_centroid(shift) - fil_centroid, x0=np.zeros(2))
    itor_per_amp = shift2d(itor_per_amp, shift)

    # Current density from current
    jtor_per_amp = itor_per_amp / (dr * dz)
    assert np.sum(jtor_per_amp) == approx(total_turns / (dr * dz), rel=1e-6, abs=1e-6)

    return jtor_per_amp


def _allocate_current_regular(
    fil_rzn: tuple[NDArray, NDArray, NDArray],
    grids: tuple[NDArray, NDArray],
    meshes: tuple[NDArray, NDArray],
) -> NDArray:
    """Assign filament current-turn density for filaments that are known to land on grid points"""
    rgrid, zgrid = grids
    dr = rgrid[1] - rgrid[0]  # [m]
    dz = zgrid[1] - zgrid[0]  # [m]
    area = dr * dz  # [m^2]

    # Map current density per amp
    jtor_per_amp = np.zeros_like(meshes[0])  # [A-turns/m^2 / A]
    for r, z, n in zip(*fil_rzn, strict=True):
        # Get indices of location of this filament
        ri = np.argmin(np.abs(rgrid - r))
        zi = np.argmin(np.abs(zgrid - z))
        # Set current density for that unit cell
        # so that the total for the cell comes out to the
        # correct total current
        jtor_per_amp[ri, zi] += n / area

    return jtor_per_amp


def _local_fields(
    jtor: NDArray, grids: tuple[NDArray, NDArray], meshes: tuple[NDArray, NDArray]
) -> tuple[NDArray, NDArray, NDArray]:
    rmesh, zmesh = meshes

    # Solve flux field
    psi = solve_flux_axisymmetric(grids, meshes, jtor)  # [Wb/A]

    # Extract flux density
    br, bz = calc_flux_density_from_flux(psi, rmesh, zmesh)  # [T/A]

    return psi, br, bz


def _make_grids_regular(rs: NDArray, zs: NDArray) -> tuple[NDArray, NDArray] | None:
    """Generate a set of regular r,z grids that span the coil winding pack centers exactly
    if possible, or None if the winding pack can't be represented exactly.

    Adds 4 grid cells of padding around the winding pack to deconflict the
    cells with nonzero current density from the boundary conditions of a
    flux solve.

    If only one unit cell is present on either axis, the grid will be
    expanded 1cm in either direction.
    """

    # Get coordinates with a unit cell
    unique_r = np.array(sorted(list(set(rs))))  # [m]
    unique_z = np.array(sorted(list(set(zs))))

    # Make sure there are enough unit cells to work with,
    # expanding dimensions if necessary
    if len(unique_r) == 1:
        r = unique_r[0]
        unique_r = [r - 1e-2, r, r + 1e-2]
    if len(unique_z) == 1:
        z = unique_z[0]
        unique_z = [z - 1e-2, z, z + 1e-2]
    if len(unique_r) < 2 or len(unique_z) < 2:
        log().error("Failed to expand grid dimensionality. This is a bug.")
        return None

    # Check if the coordinates have regular spacing,
    # which is required to support the finite difference solve
    drs = np.diff(unique_r)
    drmean = np.mean(drs)
    if np.any(np.abs(drs - drmean) / drmean > 1e-4):
        return None
    dzs = np.diff(unique_z)
    dzmean = np.mean(dzs)
    if np.any(np.abs(dzs - dzmean) / dzmean > 1e-4):
        return None

    # Extend grids by a few cells outside the winding pack
    # 7 is the true minimum; 2x 4th-order finite difference patches
    # will be stacked to extract the flux then the flux density, which means
    # 7 cells see direct interaction with boundary conditions and must not
    # have nonzero current density to produce sane results; npad = 6 produces junk outputs.
    npad = 7
    nr = len(unique_r) + 2 * npad
    nz = len(unique_z) + 2 * npad
    r_pad = npad * drmean
    z_pad = npad * dzmean
    rgrid = np.linspace(unique_r[0] - r_pad, unique_r[-1] + r_pad, nr)
    zgrid = np.linspace(unique_z[0] - z_pad, unique_z[-1] + z_pad, nz)

    # Make sure the grid doesn't cross zero.
    # If this check becomes a problem, there is an alternate strategy
    # to double resolution and spread the coil's current density mapping
    # across more than one neighboring cell, but that is too much complexity
    # to implement proactively.
    if rgrid[0] < 0.0:
        return None

    return (rgrid, zgrid)



from dataclasses import dataclass
from functools import cached_property

import numpy as np
from cfsem import (
    flux_circular_filament,
    self_inductance_annular_ring,
    self_inductance_circular_ring_wien,
    self_inductance_lyle6,
)
from interpn import MulticubicRegular
from numpy.typing import NDArray
from omas import ODS
from shapely import Polygon

from device_inductance.local import LocalFields, local_fields


@dataclass(frozen=True)
class CoilFilament:
    """
    A discretized element of an axisymmetric magnet.
    Self-inductance is calculated based on conductor geometry.
    """

    r: float
    """[m] radial location"""

    z: float
    """[m] z location"""

    n: float
    """[dimensionless] number of turns"""

    self_inductance: float
    """[H] scalar self-inductance of this filament"""


@dataclass(frozen=True)
class Coil:
    """An axisymmetric magnet, which may not have a rectangular cross-section"""

    name: str
    """This name should match the name used in the device description ODS"""

    resistance: float
    """[ohm] total effective resistance of the coil; for superconducting coils, this will be small"""

    self_inductance: float
    """[H] total scalar self-inductance of this coil"""

    filaments: list[CoilFilament]
    """Discretized circular filaments describing the coil's winding pattern"""

    @cached_property
    def grids(self) -> tuple[NDArray, NDArray]:
        """Generate a grid that spans the winding pack, landing exactly on the windings if possible."""
        return self.local_fields.grids

    @cached_property
    def meshes(self) -> tuple[NDArray, NDArray]:
        """Generate a meshgrid that spans the winding pack, landing exactly on the windings if possible."""
        return self.local_fields.meshes

    @cached_property
    def local_fields(self) -> LocalFields:
        """
        Solve the local self-field flux and flux density per amp by mapping the
        coil section to a continuous current density distribution and solving the
        continuous flux field via 4th-order finite difference, then extracting the
        B-field from the flux field via 4th-order finite difference.

        Returns:
            Structure containing local field map components and interpolators
        """
        # Polygon section representation
        # As a heuristic, treat each filament as a rectangle with side length that is
        # half the smallest distance between two sequential filaments.
        # These polygons are only used if the windings do not fall on a regular grid
        # or are too close to r=0 to allow padding the grid to preserve boundary conditions.
        drs = np.diff(self.rs)
        dzs = np.diff(self.zs)
        w = np.min(np.linalg.norm((drs, dzs), axis=0)) / 2

        polygons = [
            Polygon.from_bounds(r - w / 2, z - w / 2, r + w / 2, z + w / 2)
            for r, z in zip(self.rs, self.zs, strict=True)
        ]

        # Point-source representation
        fil_rzn = (self.rs, self.zs, self.ns)

        # Local field solve
        return local_fields(fil_rzn, polygons)

    @cached_property
    def local_field_interpolators(
        self,
    ) -> tuple[MulticubicRegular, MulticubicRegular, MulticubicRegular]:
        """Build interpolators over the solved local fields, if available"""
        return (
            self.local_fields.psi_per_amp_interpolator,
            self.local_fields.br_per_amp_interpolator,
            self.local_fields.bz_per_amp_interpolator,
        )

    @cached_property
    def extent(self) -> tuple[float, float, float, float]:
        """[m] rmin, rmax, zmin, zmax extent of filament centers"""
        r = [f.r for f in self.filaments]
        z = [f.z for f in self.filaments]
        return min(r), max(r), min(z), max(z)

    @cached_property
    def rs(self) -> NDArray:
        """[m] Filament r-coordinates"""
        return np.array([f.r for f in self.filaments])

    @cached_property
    def zs(self) -> NDArray:
        """[m] Filament z-coordinates"""
        return np.array([f.z for f in self.filaments])

    @cached_property
    def ns(self) -> NDArray:
        """[dimensionless] Filament number of turns"""
        return np.array([f.n for f in self.filaments])


def _extract_coils(description: ODS) -> list[Coil]:
    """
    Extract coil filamentization and, while full geometric info is available,
    calculate self-inductance of individual filaments and of the coil as a whole.

    Because some coils are not of rectangular cross-section, and even some coils of
    rectangular cross-section do not have evenly-distributed number of turns between
    different elements, the coil's self-inductance is calculated by using the singularity
    method to calculate the mutual inductance between each pair of elements, then replacing
    the singular self-field terms with the corresponding element's estimated self-inductance.

    An approximate calc is used for the self-inductance of individual elements, which
    can't use the singularity method. The existing method adequately handles
    rectangular-section elements, but approximates each non-rectangular element as a
    square section with the same area, which may introduce some error when handling elements
    of circular, annular, or other cross-sectional geometry. More detailed handling can be
    added later, giving higher resolution for such cases.

    Args:
        description: Device geometric info in the format produced by device_description

    Raises:
        ValueError: If an un-handled type of coil element cross-sectional geometry is encountered

    Returns:
        A list of coil objects, populated with reduced geometric info and estimated self-inductances.
    """

    coils: list[Coil] = []
    for ods_coil in description["pf_active.coil"].values():
        coil_name = ods_coil["name"]
        resistance = ods_coil["resistance"]
        coil_filaments: list[CoilFilament] = []

        # Process individual elements
        for coil_elem in ods_coil["element"].values():
            geom_type = coil_elem["geometry.geometry_type"]
            turns_with_sign = coil_elem["turns_with_sign"]  # [dimensionless]
            n = abs(turns_with_sign)  # Circuit definition is responsible for sign

            # Approximate the self-inductance of the individual elements
            # as rectangular sections, solid rings, or annular rings
            # depending on geometry type id.
            if geom_type == 5:
                # Annular section (or, with zero inner radius, solid circular)
                r = coil_elem["geometry.annulus.r"]  # [m]
                z = coil_elem["geometry.annulus.z"]  # [m]
                ri = coil_elem["geometry.annulus.radius_inner"]  # [m]
                ro = coil_elem["geometry.annulus.radius_outer"]  # [m]

                if ri > 1e-4:
                    elem_self_inductance = self_inductance_annular_ring(r, ri, ro)  # [H]
                else:
                    # Use solid ring calc for small inner radius to avoid div/0
                    elem_self_inductance = self_inductance_circular_ring_wien(r, ro)  # [H]

            elif geom_type == 2:
                # Solid rectangular section
                r = coil_elem["geometry.rectangle.r"]  # [m]
                z = coil_elem["geometry.rectangle.z"]  # [m]
                w = coil_elem["geometry.rectangle.width"]  # [m]
                h = coil_elem["geometry.rectangle.height"]  # [m]

                elem_self_inductance = self_inductance_lyle6(r, w, h, n)  # [H]
            else:
                raise ValueError(f"Unhandled coil element geometry type: {geom_type}")

            # Store the parts we need for calculating mutual inductances
            coil_filaments.append(CoilFilament(r=r, z=z, n=n, self_inductance=float(elem_self_inductance)))

        # Calculate self-inductance of the whole coil
        coil_self_inductance = 0.0  # [H]
        elem_rs = np.array([x.r for x in coil_filaments])  # [m]
        elem_zs = np.array([x.z for x in coil_filaments])  # [m]
        elem_ns = np.array([x.n for x in coil_filaments])  # [dimensionless]
        elem_self_inductances = np.array([x.self_inductance for x in coil_filaments])  # [H]
        nelem = len(elem_rs)
        for i in range(nelem):
            this_r = np.array([elem_rs[i]])  # [m]
            this_z = np.array([elem_zs[i]])  # [m]
            this_n = np.array([elem_ns[i]])  # [dimensionless]

            # Use one-to-many flux calc for speed
            contribs = elem_ns * flux_circular_filament(
                ifil=this_n,  # Unit current multiplied by number of turns
                rfil=this_r,
                zfil=this_z,
                rprime=np.array(elem_rs),
                zprime=np.array(elem_zs),
            )
            # Use precalcualted self-contribution which is otherwise singular and handled separately
            contribs[i] = elem_self_inductances[i]

            # Total contribution to coil self-inductance
            turn_contrib = np.sum(contribs)
            coil_self_inductance += float(turn_contrib)

        coil = Coil(
            name=coil_name,
            resistance=resistance,  # [ohm]
            self_inductance=coil_self_inductance,  # [H]
            filaments=coil_filaments,
        )
        coils.append(coil)

    return coils



"""Grid input definitions"""

GridSpec = tuple[float, int, float, int]
"""[m] rmin, nr, zmin, nz exact specification of regular grid,
as an alternative to approximate minimum extent"""


Extent = tuple[float, float, float, float]
"""[m] rmin, rmax, zmin, zmax rectangular bounds"""


Resolution = tuple[float, float]
"""[m] dr, dz spatial resolution of computational grid"""
